from __future__ import annotations

import base64
import json
import os
from pathlib import Path
from typing import Literal, TypedDict

from . import _utils


# This is the same as the FileContentJson type in TypeScript.
class FileContentJson(TypedDict):
    name: str
    content: str
    type: Literal["text", "binary"]


class AppInfo(TypedDict):
    appdir: str
    subdir: str
    files: list[FileContentJson]


# =============================================================================
def read_app_files(appdir: Path, destdir: Path) -> list[FileContentJson]:
    """
    Load files for a Shiny application.

    Parameters
    ----------
    appdir : str
       Directory containing the application.

    destdir : str
       Destination directory. This is used only to avoid adding shinylive assets when
       they are in a subdir of the application.
    """
    app_files: list[FileContentJson] = []
    # Recursively iterate over files in app directory, and collect the files into
    # app_files data structure.
    exclude_names = {"__pycache__", "venv", ".venv"}
    for root, dirs, files in os.walk(appdir, topdown=True):
        root = Path(root)

        if _utils.is_relative_to(Path(root), destdir):
            # In case destdir is inside of the appdir, don't copy those files.
            continue

        dirs[:] = [d for d in dirs if not d.startswith(".")]
        dirs[:] = set(dirs) - exclude_names
        rel_dir = root.relative_to(appdir)
        files = [f for f in files if not f.startswith(".")]
        files = [f for f in files if f not in exclude_names]
        files.sort()

        # Move app.py to first in list.
        if "app.py" in files:
            app_py_idx = files.index("app.py")
            files.insert(0, files.pop(app_py_idx))

        # Add the file to the app_files list.
        for filename in files:
            if rel_dir == ".":  # pyright: ignore[reportUnnecessaryComparison]
                output_filename = filename
            else:
                output_filename = str(rel_dir / filename)

            if filename == "shinylive.js":
                print(
                    f"Warning: Found shinylive.js in source directory '{appdir}/{rel_dir}'. Are you including a shinylive distribution in your app?"
                )

            type: Literal["text", "binary"] = "text"
            try:
                with open(root / filename, "r") as f:
                    file_content = f.read()
                    type = "text"
            except UnicodeDecodeError:
                # If text failed, try binary.
                with open(root / filename, "rb") as f:
                    file_content_bin = f.read()
                    file_content = base64.b64encode(file_content_bin).decode("utf-8")
                    type = "binary"

            app_files.append(
                {
                    "name": output_filename,
                    "content": file_content,
                    "type": type,
                }
            )

    return app_files


def write_app_json(app_info: AppInfo, destdir: Path, html_source_dir: Path) -> None:
    """
    Write index.html, edit/index.html, and app.json for an application in the destdir.
    """
    app_destdir = destdir / app_info["subdir"]

    # For a subdir like a/b/c, this will be ../../../
    subdir_inverse = "/".join([".."] * _utils.path_length(app_info["subdir"]))
    if subdir_inverse != "":
        subdir_inverse += "/"

    if not app_destdir.exists():
        app_destdir.mkdir()

    _utils.copy_file_and_substitute(
        src=html_source_dir / "index.html",
        dest=app_destdir / "index.html",
        replacements=(
            ("{{REL_PATH}}", subdir_inverse),
            ("{{APP_ENGINE}}", "python"),
        ),
    )

    editor_destdir = app_destdir / "edit"
    if not editor_destdir.exists():
        editor_destdir.mkdir()
    _utils.copy_file_and_substitute(
        src=html_source_dir / "edit" / "index.html",
        dest=(editor_destdir / "index.html"),
        replacements=(
            ("{{REL_PATH}}", subdir_inverse),
            ("{{APP_ENGINE}}", "python"),
        ),
    )

    app_json_output_file = app_destdir / "app.json"

    print("Writing " + str(app_json_output_file), end="")
    json.dump(app_info["files"], open(app_json_output_file, "w"))
    print(":", app_json_output_file.stat().st_size, "bytes")



from __future__ import annotations

import os
import sys
from pathlib import Path

from . import _deps, _utils
from ._app_json import AppInfo, read_app_files, write_app_json
from ._assets import shinylive_assets_dir


def export(
    appdir: str | Path,
    destdir: str | Path,
    *,
    subdir: str | Path | None = None,
    verbose: bool = False,
    full_shinylive: bool = False,
):
    def verbose_print(*args: object) -> None:
        if verbose:
            print(*args)

    appdir = Path(appdir)
    destdir = Path(destdir)

    if not (appdir / "app.py").exists():
        raise ValueError(f"Directory {appdir}/ must contain a file named app.py.")

    if subdir is None:
        subdir = ""
    subdir = Path(subdir)
    if subdir.is_absolute():
        raise ValueError(
            f"subdir {subdir} is absolute, but only relative paths are allowed."
        )

    if not destdir.exists():
        print(f"Creating {destdir}/", file=sys.stderr)
        destdir.mkdir()

    copy_fn = _utils.create_copy_fn(overwrite=False, verbose_print=verbose_print)

    assets_dir = Path(shinylive_assets_dir())

    # =========================================================================
    # Copy the base dependencies for shinylive/ distribution. This does not include the
    # Python package files.
    # =========================================================================
    print(
        f"Copying base Shinylive files from {assets_dir}/ to {destdir}/",
        file=sys.stderr,
    )

    base_files = _deps.shinylive_common_files(
        # Do not include r-only support files
        asset_type=("base", "python"),
    )
    for file in base_files:
        src_path = assets_dir / file
        dest_path = destdir / Path(file)

        if not dest_path.parent.exists():
            os.makedirs(dest_path.parent)

        copy_fn(src_path, dest_path)

    # =========================================================================
    # Load each app's contents into a list[FileContentJson]
    # =========================================================================
    app_info: AppInfo = {
        "appdir": str(appdir),
        "subdir": str(subdir),
        "files": read_app_files(appdir, destdir),
    }

    # =========================================================================
    # Copy dependencies from shinylive/pyodide/
    # =========================================================================
    if full_shinylive:
        package_files = _utils.listdir_recursive(assets_dir / "shinylive" / "pyodide")
        # Some of the files in this dir are base files; don't copy them.
        package_files = [
            file
            for file in package_files
            if os.path.join("shinylive", "pyodide", file) not in base_files
        ]

    else:
        deps = _deps.base_package_deps() + _deps.find_package_deps(app_info["files"])

        package_files: list[str] = [dep["file_name"] for dep in deps]

        print(
            f"Copying imported packages from {assets_dir}/shinylive/pyodide/ to {destdir}/shinylive/pyodide/",
            file=sys.stderr,
        )
        verbose_print(" ", ", ".join(package_files))

    for filename in package_files:
        src_path = assets_dir / "shinylive" / "pyodide" / filename
        dest_path = destdir / "shinylive" / "pyodide" / filename
        if not dest_path.parent.exists():
            os.makedirs(dest_path.parent)

        copy_fn(src_path, dest_path)

    # =========================================================================
    # For each app, write the index.html, edit/index.html, and app.json in
    # destdir/subdir.
    # =========================================================================
    write_app_json(
        app_info,
        destdir,
        html_source_dir=Path(shinylive_assets_dir()) / "export_template",
    )

    print(
        "\nRun the following to serve the app:\n"
        f"  python3 -m http.server --directory {destdir} --bind localhost 8008",
        file=sys.stderr,
    )



from __future__ import annotations

import ast
import copy
import functools
import json
import os
import sys
from pathlib import Path
from textwrap import dedent
from typing import Callable, Iterable, Literal, Optional, Tuple

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict

from ._app_json import FileContentJson
from ._assets import ensure_shinylive_assets, repodata_json_file, shinylive_assets_dir
from ._version import SHINYLIVE_ASSETS_VERSION

# Files in Pyodide that should always be included.
BASE_PYODIDE_FILES = {
    "pyodide_py.tar",
    "pyodide.asm.js",
    "pyodide.asm.data",
    "pyodide.asm.wasm",
    "repodata.json",
}

# Packages that should always be included in a Shinylive deployment.
BASE_PYODIDE_PACKAGES = {"distutils", "micropip", "ssl"}
AssetType = Literal["base", "python", "r"]


# =============================================================================
# Data structures used in pyodide/repodata.json
# =============================================================================
# Note: This block of code is copied from /scripts/pyodide_packages.py
class PyodidePackageInfo(TypedDict):
    name: str
    version: str
    file_name: str
    install_dir: Literal["lib", "site"]
    sha256: str
    depends: list[str]
    imports: list[str]
    unvendored_tests: NotRequired[bool]


# The package information structure used by Pyodide's repodata.json.
class PyodideRepodataFile(TypedDict):
    info: dict[str, str]
    packages: dict[str, PyodidePackageInfo]


# =============================================================================
# HTML Dependency types
# =============================================================================
class HtmlDepItem(TypedDict):
    name: str
    path: str
    attribs: NotRequired[dict[str, str]]


class HtmlDepServiceworkerItem(TypedDict):
    source: str
    destination: str


class QuartoHtmlDependency(TypedDict):
    name: str
    version: NotRequired[str]
    scripts: NotRequired[list[str | HtmlDepItem]]
    stylesheets: NotRequired[list[str | HtmlDepItem]]
    resources: NotRequired[list[HtmlDepItem]]
    meta: NotRequired[dict[str, str]]
    serviceworkers: NotRequired[list[HtmlDepServiceworkerItem]]


# =============================================================================
# Conversion functions
# =============================================================================
def _dep_names_to_pyodide_pkg_infos(
    dep_names: Iterable[str],
) -> list[PyodidePackageInfo]:
    repodata = _pyodide_repodata()
    pkg_infos: list[PyodidePackageInfo] = [
        copy.deepcopy(repodata["packages"][dep_name]) for dep_name in dep_names
    ]
    return pkg_infos


def _pyodide_pkg_info_to_quarto_html_dep_item(
    pkg: PyodidePackageInfo,
) -> HtmlDepItem:
    """
    Convert a PyodidePackageInfo object to a HtmlDepItem object.
    """

    assets_dir = shinylive_assets_dir()

    return {
        "name": os.path.join("shinylive", "pyodide", pkg["file_name"]),
        "path": os.path.join(assets_dir, "shinylive", "pyodide", pkg["file_name"]),
    }


def _pyodide_pkg_infos_to_quarto_html_dep_items(
    pkgs: list[PyodidePackageInfo],
) -> list[HtmlDepItem]:
    return [_pyodide_pkg_info_to_quarto_html_dep_item(pkg) for pkg in pkgs]


# =============================================================================
# Shinylive base dependencies
# =============================================================================
def shinylive_base_deps_htmldep(
    sw_dir: Optional[str] = None,
    *,
    asset_type: Tuple[AssetType] = ("base",),
) -> list[QuartoHtmlDependency]:
    return [
        _serviceworker_dep(sw_dir),
        _shinylive_common_dep_htmldep(asset_type=asset_type),
    ]


def shinylive_python_resources() -> list[HtmlDepItem]:
    ret: list[HtmlDepItem] = []

    # Add python support file
    rel_paths = shinylive_common_files(asset_type=("python",))
    assets_dir = shinylive_assets_dir()
    ret.extend(
        [
            {"name": rel_path, "path": os.path.join(assets_dir, rel_path)}
            for rel_path in rel_paths
        ]
    )

    # Add base python packages as resources
    ret.extend(base_package_deps_htmldepitems())

    return ret


# Not used in practice!
def shinylive_r_resources() -> list[HtmlDepItem]:
    rel_paths = shinylive_common_files(asset_type=("r",))
    assets_dir = shinylive_assets_dir()
    return [
        {"name": rel_path, "path": os.path.join(assets_dir, rel_path)}
        for rel_path in rel_paths
    ]


# =============================================================================
# Common dependencies
# =============================================================================
def _shinylive_common_dep_htmldep(
    *,
    asset_type: Tuple[AssetType, ...] = (
        "base",
        "python",
    ),
) -> QuartoHtmlDependency:
    """
    Return an HTML dependency object consisting of files that are base dependencies; in
    other words, the files that are always included in a Shinylive deployment.
    """
    assets_dir = shinylive_assets_dir()

    # First, get the list of base files.
    base_files = shinylive_common_files(asset_type=asset_type)

    # Next, categorize the base files into scripts, stylesheets, and resources.
    scripts: list[str | HtmlDepItem] = []
    stylesheets: list[str | HtmlDepItem] = []
    resources: list[HtmlDepItem] = []

    for file in base_files:
        if os.path.basename(file) in [
            "load-shinylive-sw.js",
            "run-python-blocks.js",
        ]:
            script_item: HtmlDepItem = {
                "name": file,
                "path": os.path.join(assets_dir, file),
                "attribs": {"type": "module"},
            }

            scripts.append(script_item)

        if os.path.basename(file) in [
            "shinylive.css",
        ]:
            stylesheets.append(
                {
                    "name": file,
                    "path": os.path.join(assets_dir, file),
                }
            )
        else:
            resources.append(
                {
                    "name": file,
                    "path": os.path.join(assets_dir, file),
                }
            )

    # Sort scripts so that load-shinylive-sw.js is first, and run-python-blocks.js is
    # last.
    def scripts_sort_fun(x: str | HtmlDepItem) -> int:
        if isinstance(x, str):
            filename = x
        else:
            filename = os.path.basename(x["name"])

        if filename == "load-shinylive-sw.js":
            return 0
        elif filename == "run-python-blocks.js":
            return 2
        else:
            return 1

    scripts.sort(key=scripts_sort_fun)

    return {
        "name": "shinylive",
        "version": SHINYLIVE_ASSETS_VERSION,
        "scripts": scripts,
        "stylesheets": stylesheets,
        "resources": resources,
    }


def shinylive_common_files(
    *,
    asset_type: Tuple[AssetType, ...],
) -> list[str]:
    """
    Return a list of asset files for Python, and/or R, and/or language-agnostic (base) dependencies
    """
    ensure_shinylive_assets()

    has_base = "base" in asset_type
    has_python = "python" in asset_type
    has_r = "r" in asset_type

    base_files: list[str] = []
    for root, dirs, files in os.walk(shinylive_assets_dir()):
        root = Path(root)
        rel_root = root.relative_to(shinylive_assets_dir())
        if rel_root == Path("."):
            dirs.remove("scripts")
            dirs.remove("export_template")
            if not has_base:
                # No files to add here
                files = []
        elif rel_root == Path("shinylive"):
            files.remove("examples.json")
            # Remove webr folder as it is only needed for R support
            if not has_r:
                dirs.remove("webr")
            if not has_python:
                dirs.remove("pyodide")
                dirs.remove("pyright")
            if not has_base:
                # No files to add here
                files = []
        elif rel_root == Path("shinylive/pyodide"):
            dirs.remove("fonts")
            files[:] = BASE_PYODIDE_FILES

        for file in files:
            if file.startswith("."):
                continue
            base_files.append(str(rel_root / file))

    return base_files


def _serviceworker_dep(sw_dir: Optional[str] = None) -> QuartoHtmlDependency:
    dep: QuartoHtmlDependency = {
        "name": "shinylive-serviceworker",
        "version": SHINYLIVE_ASSETS_VERSION,
        "serviceworkers": [
            {
                "source": shinylive_assets_dir() + "/shinylive-sw.js",
                "destination": "/shinylive-sw.js",
            }
        ],
    }

    if sw_dir is not None:
        # Add meta tag to tell load-shinylive-sw.js where to find shinylive-sw.js.
        dep["meta"] = {"shinylive:serviceworker_dir": sw_dir}

    return dep


# =============================================================================
# Find which packages are used by a Shiny application
# =============================================================================
def shinylive_app_resources(
    json_file: Optional[str | Path],
    json_content: Optional[str],
) -> list[HtmlDepItem]:
    """
    Find package dependencies from an app.json file, and return as a list of
    QuartoHtmlDependency objects.

    Requires either `json_file` or `json_content`, but not both.
    """

    if (json_file is None and json_content is None) or (
        json_file is not None and json_content is not None
    ):
        raise RuntimeError("Must provide either `json_file` or `json_content`.")

    file_contents: list[FileContentJson] = []

    if json_file is not None:
        json_file = Path(json_file)
        with open(json_file) as f:
            file_contents = json.load(f)

    if json_content is not None:
        file_contents = json.loads(json_content)

    pkg_infos = find_package_deps(file_contents)
    deps = _pyodide_pkg_infos_to_quarto_html_dep_items(pkg_infos)
    return deps


def find_package_deps(
    app_contents: list[FileContentJson],
    verbose_print: Callable[..., None] = lambda *args: None,
) -> list[PyodidePackageInfo]:
    """
    Find package dependencies from the contents of an app.json file. The returned data
    structure is a list of PyodidePackageInfo objects.
    """

    imports: set[str] = _find_import_app_contents(app_contents)

    # TODO: Need to also add in requirements.txt, and find dependencies of those
    # packages, in case any of those dependencies are included as part of pyodide.
    verbose_print("Imports detected in app:\n ", ", ".join(sorted(imports)))

    dep_names = _find_recursive_deps(imports, verbose_print)
    pkg_infos = _dep_names_to_pyodide_pkg_infos(dep_names)

    return pkg_infos


def base_package_deps_htmldepitems() -> list[HtmlDepItem]:
    """
    Return list of python packages that should be included in all python Shinylive
    deployments. The returned data structure is a list of HtmlDepItem objects
    representing PyodidePackageInfo objects.
    """
    pkg_infos = base_package_deps()
    deps = _pyodide_pkg_infos_to_quarto_html_dep_items(pkg_infos)

    return deps


def base_package_deps() -> list[PyodidePackageInfo]:
    """
    Return list of python packages that should be included in all python Shinylive
    deployments. The returned data structure is a list of PyodidePackageInfo objects.
    """
    dep_names = _find_recursive_deps(BASE_PYODIDE_PACKAGES)
    pkg_infos = _dep_names_to_pyodide_pkg_infos(dep_names)

    return pkg_infos


# =============================================================================
# Internal functions
# =============================================================================
def _find_recursive_deps(
    pkgs: Iterable[str],
    verbose_print: Callable[..., None] = lambda *args: None,
) -> list[str]:
    """
    Given a list of packages, recursively find all dependencies that are contained in
    repodata.json. This returns a list of all dependencies, including the original
    packages passed in.
    """
    repodata = _pyodide_repodata()
    deps = list(pkgs)
    i = 0
    while i < len(deps):
        dep = deps[i]
        if dep not in repodata["packages"]:
            # TODO: Need to distinguish between built-in packages and external ones in
            # requirements.txt.
            verbose_print(
                f"  {dep} not in repodata.json. Assuming it is in base Pyodide or in requirements.txt."
            )
            deps.remove(dep)
            continue

        dep_deps = set(repodata["packages"][dep]["depends"])
        new_deps = dep_deps.difference(deps)
        deps.extend(new_deps)
        i += 1

    return deps


def _dep_name_to_dep_file(dep_name: str) -> str:
    """
    Given the name of a dependency, like "pandas", return the name of the .whl file,
    like "pandas-1.4.2-cp310-cp310-emscripten_3_1_14_wasm32.whl".
    """
    repodata = _pyodide_repodata()
    return repodata["packages"][dep_name]["file_name"]


def _dep_names_to_dep_files(dep_names: list[str]) -> list[str]:
    """
    Given a list of dependency names, like ["pandas"], return a list with the names of
    corresponding .whl files (from data in repodata.json), like
    ["pandas-1.4.2-cp310-cp310-emscripten_3_1_14_wasm32.whl"].
    """
    repodata = _pyodide_repodata()
    dep_files = [repodata["packages"][x]["file_name"] for x in dep_names]
    return dep_files


def _find_import_app_contents(app_contents: list[FileContentJson]) -> set[str]:
    """
    Given an app.json file, find packages that are imported.
    """
    imports: set[str] = set()
    for file_content in app_contents:
        if not file_content["name"].endswith(".py"):
            continue

        imports = imports.union(_find_imports(file_content["content"]))

    # Note that at this point, the imports are module names, like "cv2", but these can
    # sometimes differ from the package names, like "opencv-python". We need to map from
    # module names to package names.
    packages = [module_to_package(x) for x in imports]
    packages = [x for x in packages if x is not None]

    return set(packages)


def module_to_package(module: str) -> str | None:
    """
    Given a module name, like "cv2", return the corresponding package name, like
    "opencv-python". If not found, return None.
    """
    module_to_package = _module_to_package_mappings()
    if module in module_to_package:
        return module_to_package[module]
    else:
        return None


@functools.lru_cache
def _module_to_package_mappings() -> dict[str, str]:
    """
    Return a dictionary that maps module names to package names. This is needed because
    sometimes the module name and package name are different. For example, the module
    name is "cv2", but the package name is "opencv-python".
    """
    repodata = _pyodide_repodata()
    module_to_package: dict[str, str] = {}
    for pkg_name, pkg_info in repodata["packages"].items():
        modules = pkg_info["imports"]
        for module in modules:
            module_to_package[module] = pkg_name

    return module_to_package


@functools.lru_cache
def _pyodide_repodata() -> PyodideRepodataFile:
    """
    Read in the Pyodide repodata.json file and return the contents. The result is
    cached, so if the file changes, it won't register until the Python session is
    restarted.
    """
    with open(repodata_json_file(), "r") as f:
        return json.load(f)


# From pyodide._base.find_imports
def _find_imports(source: str) -> list[str]:
    """
    Finds the imports in a Python source code string

    Parameters
    ----------
    source : str
       The Python source code to inspect for imports.

    Returns
    -------
    ``list[str]``
        A list of module names that are imported in ``source``. If ``source`` is not
        syntactically correct Python code (after dedenting), returns an empty list.

    Examples
    --------
    >>> from pyodide import find_imports
    >>> source = "import numpy as np; import scipy.stats"
    >>> find_imports(source)
    ['numpy', 'scipy']
    """
    # handle mis-indented input from multi-line strings
    source = dedent(source)

    try:
        mod = ast.parse(source)
    except SyntaxError:
        return []
    imports: set[str] = set()
    for node in ast.walk(mod):
        if isinstance(node, ast.Import):
            for name in node.names:
                node_name = name.name
                imports.add(node_name.split(".")[0])
        elif isinstance(node, ast.ImportFrom):
            module_name = node.module
            if module_name is None:
                continue
            imports.add(module_name.split(".")[0])
    return list(sorted(imports))



"""A package for packaging Shiny applications that run on Python in the browser."""

from ._url import ShinyliveApp, url_decode, url_encode
from ._version import SHINYLIVE_PACKAGE_VERSION

__version__ = SHINYLIVE_PACKAGE_VERSION

__all__ = ("ShinyliveApp", "url_decode", "url_encode")



from __future__ import annotations

import base64
import copy
import json
import os
import re
import sys
from pathlib import Path
from typing import Any, Literal, Optional, Sequence, cast
from urllib.parse import urlparse

# Even though TypedDict is available in Python 3.8, because it's used with NotRequired,
# they should both come from the same typing module.
# https://peps.python.org/pep-0655/#usage-in-python-3-11
if sys.version_info >= (3, 11):
    from typing import NotRequired, TypedDict
else:
    from typing_extensions import NotRequired, TypedDict


class FileContentJson(TypedDict):
    name: str
    content: str
    type: NotRequired[Literal["text", "binary"]]


SHINYLIVE_CODE_TEMPLATE = """
```{{shinylive-{language}}}
#| standalone: true
#| components: [{components}]
#| layout: {layout}
#| viewerHeight: {viewer_height}

{contents}
```
"""


class ShinyliveApp:
    """
    Create an instance of a Shiny App for use with shinylive.io.

    Parameters
    ----------
    bundle
        The file bundle of the Shinylive application. This should be a list of files
        as a dictionary of "name", "content" and optionally `"type": "binary"` for
        binary file types. (`"type": "text"` is the default and can be omitted.)
    language
        The language of the application, or None to autodetect the language. Defaults
        to None.
    mode
        The mode of the application, used when creating a shinylive.io URL. Accepted
        values are either "editor" or "app"; defaults to "editor".
    header
        Whether to include a header bar in the UI when creating a shinylive.io URL. This
        is used only if ``mode`` is "app". Defaults to True.
    host
        The host URL of the Shinylive application. Defaults to "https://shinylive.io".
    """

    def __init__(
        self,
        bundle: list[FileContentJson],
        language: Optional[Literal["py", "r"]],
        mode: Literal["editor", "app"] = "editor",
        header: bool = True,
        host: str = "https://shinylive.io",
    ):
        self._bundle = bundle
        if language is None:
            lang = detect_app_language(bundle[0]["content"])
        else:
            if language not in ["py", "r"]:
                raise ValueError(
                    f"Invalid language '{language}', must be either 'py' or 'r'."
                )
            lang = language

        if mode not in ["editor", "app"]:
            raise ValueError(
                f"Invalid mode '{mode}', must be either 'editor' or 'app'."
            )

        if not isinstance(header, bool):
            raise ValueError(
                f"Invalid header '{header}', must be either True or False."
            )

        self.mode: Literal["editor", "app"] = mode
        self.header: bool = header
        self.host: str = host
        self._language: Literal["py", "r"] = lang
        self._app_path: Optional[Path] = None
        self._root_dir: Optional[Path] = None

    @classmethod
    def from_local(
        cls,
        app: str | Path,
        files: Optional[str | Path | Sequence[str | Path]] = None,
        language: Optional[Literal["py", "r"]] = None,
        **kwargs: Any,
    ) -> ShinyliveApp:
        """
        Create an instance of a Shiny App from local files for use with shinylive.io.

        Parameters
        ----------
        app
            The main app file of the Shinylive application. This file should be a Python
            `app.py` or an R `app.R`, `ui.R`, or `server.R` file. This file will be
            renamed `app.py` or `app.R` for shinylive, unless it's named `ui.R` or
            `server.R`.
        files
            File(s) or directory path(s) to include in the application. On shinylive,
            these files will be stored relative to the main `app` file.
        language
            The language of the application, or None to autodetect the language. Defaults
            to None.
        """
        if language is None:
            language = detect_app_language(app)
        elif language not in ["py", "r"]:
            raise ValueError(
                f"Language '{language}' is not supported. Please specify one of 'py' or 'r'."
            )

        self = cls([], language=language, **kwargs)

        self._app_path = Path(app)
        self._root_dir = self._app_path.parent
        app_fc = read_file(app, self._root_dir)

        # if the app is not named either `ui.R` or `server.R`, then make it app.py or app.R
        if app_fc["name"] not in ["ui.R", "server.R"]:
            app_fc["name"] = f"app.{'py' if self._language == 'py' else 'R'}"

        self._bundle.append(app_fc)
        self.add_files(files)
        return self

    @classmethod
    def from_text(
        cls,
        app_text: str,
        files: Optional[str | Path | Sequence[str | Path]] = None,
        language: Optional[Literal["py", "r"]] = None,
        root_dir: Optional[str | Path] = None,
        **kwargs: Any,
    ) -> ShinyliveApp:
        """
        Create an instance of a Shiny App from a string containing the `app.py` or `app.R`
        file contents for use with shinylive.io.

        Parameters
        ----------
        app_code
            The text contents of the main app file for the Shinylive application. This file
            will be renamed `app.py` or `app.R` for shinylive.
        files
            File(s) or directory path(s) to include in the application. On shinylive,
            these files will be stored relative to the main `app` file.
        language
            The language of the application, or None to autodetect the language. Defaults
            to None.
        root_dir
            The root directory of the application,used to determine the relative
            path of supporting files to the main ``app`` file. Defaults to ``None``, meaning
            that additional files are added in a flattened structure.
        """
        if language is None:
            language = detect_app_language(app_text)
        elif language not in ["py", "r"]:
            raise ValueError(
                f"Language '{language}' is not supported. Please specify one of 'py' or 'r'."
            )

        default_app_file = f"app.{'py' if language == 'py' else 'R'}"
        app_fc = {default_app_file: app_text}

        self = cls([], language=language, **kwargs)
        self._root_dir = Path(root_dir) if root_dir is not None else None

        return self.add_file_contents(app_fc).add_files(files)

    @classmethod
    def from_url(cls, url: str) -> ShinyliveApp:
        """
        Create an instance of a Shiny App from a shinylive.io URL.

        Parameters
        ----------
        url
            The shinylive.io URL to decode.
        """

        url = url.strip()
        bundle = bundle_from_url(url)
        language = "r" if "shinylive.io/r/" in url else "py"
        mode = "app" if f"{language}/app/" in url else "editor"
        header = False if "h=0" in url else True
        scheme, netloc, *_ = urlparse(url)

        return cls(
            bundle,
            language=language,
            mode=mode,
            header=header,
            host=f"{scheme}://{netloc}",
        )

    def __str__(self) -> str:
        return self.to_url()

    def to_url(
        self,
        mode: Optional[Literal["editor", "app"]] = None,
        header: Optional[bool] = None,
        host: Optional[str] = None,
    ) -> str:
        """
        Get the URL of the Shinylive application.

        Parameters
        ----------
        mode
            The mode of the application, either "editor" or "app". Defaults to the
            current mode.
        header
            Whether to include a header bar in the UI. This is used only if ``mode`` is
            "app". Defaults to the current header value.
        host
            The host URL of the Shinylive application. Defaults to the current host URL,
            which is typically ``"https://shinylive.io"``.

        Returns
        -------
        str
            The URL of the Shinylive application.
        """
        mode = mode or self.mode
        header = header if header is not None else self.header

        if mode not in ["editor", "app"]:
            raise ValueError(
                f"Invalid mode '{mode}', must be either 'editor' or 'app'."
            )

        file_lz = lzstring_file_bundle(self._bundle)

        base = (host or self.host).rstrip("/")

        h = "h=0&" if not header and mode == "app" else ""

        return f"{base}/{self._language}/{mode}/#{h}code={file_lz}"

    def view(self) -> None:
        """
        Open the Shinylive application in a browser.
        """
        import webbrowser

        webbrowser.open(self.to_url())

    def to_chunk_contents(self) -> str:
        """
        Create the contents of a Shinylive chunk based on the files in the app. This
        output does not include the Shinylive chunk header or options.

        Returns
        -------
        str
            The contents of the Shinylive chunk.
        """
        lines: list[str] = []
        for file in self._bundle:
            lines.append(f"## file: {file['name']}")
            if "type" in file and file["type"] == "binary":
                lines.append("## type: binary")
            lines.append(
                file["content"].encode("utf-8", errors="ignore").decode("utf-8")
            )
            lines.append("")

        return "\n".join(lines)

    def to_chunk(
        self,
        components: Sequence[Literal["editor", "viewer"]] = ("editor", "viewer"),
        layout: Literal["horizontal", "vertical"] = "horizontal",
        viewer_height: int = 500,
    ) -> str:
        """
        Create a Shinylive chunk based on the files in the app for use in a Quarto
        web document.

        Parameters
        ----------
        components
            Which components to include in the chunk. Defaults to both "editor" and
            "viewer".
        layout
            The layout of the components, either "horizontal" or "vertical". Defaults
            to "horizontal".
        viewer_height
            The height of the viewer component in pixels. Defaults to 500.

        Returns
        -------
        str
            The full Shinylive chunk, including the chunk header and options.
        """
        if layout not in ["horizontal", "vertical"]:
            raise ValueError(
                f"Invalid layout '{layout}', must be either 'horizontal' or 'vertical'."
            )

        if not isinstance(components, Sequence) or not all(
            component in ["editor", "viewer"] for component in components
        ):
            raise ValueError(
                f"Invalid components '{components}', must be a list or tuple of 'editor' or 'viewer'."
            )

        return SHINYLIVE_CODE_TEMPLATE.format(
            language="python" if self._language == "py" else "r",
            components=", ".join(components),
            layout=layout,
            viewer_height=viewer_height,
            contents=self.to_chunk_contents(),
        )

    def to_json(self, **kwargs: Any) -> str:
        """
        Get the JSON representation of the Shinylive application.

        Parameters
        ----------
        kwargs
            Keyword arguments passed to ``json.dumps``.

        Returns
        -------
        str
            The JSON representation of the Shinylive application.
        """
        return json.dumps(self._bundle, **kwargs)

    def write_files(self, dest: str | Path) -> Path:
        """
        Write the files in the Shinylive application to a directory.

        Parameters
        ----------
        dest
            The directory to write the files to.

        Returns
        -------
        Path
            The directory that the files were written to.
        """
        out_dir = Path(dest)
        out_dir.mkdir(parents=True, exist_ok=True)
        for file in self._bundle:
            if "type" in file and file["type"] == "binary":
                import base64

                with open(out_dir / file["name"], "wb") as f_out:
                    f_out.write(base64.b64decode(file["content"]))
            else:
                with open(out_dir / file["name"], "w") as f_out:
                    f_out.write(
                        file["content"].encode("utf-8", errors="ignore").decode("utf-8")
                    )

        return out_dir

    def add_files(
        self,
        files: Optional[str | Path | Sequence[str | Path]] = None,
    ) -> ShinyliveApp:
        """
        Add files to the Shinylive application. For more control over the file name,
        use the ``add_file`` method.

        Parameters
        ----------
        files
            File(s) or directory path(s) to include in the application. On shinylive,
            these files will be stored relative to the main ``app`` file. Use the
            ``add_file`` method to add a single file if you need to rename the files.
            In app bundles created from local files, added files will be stored relative
            to the location of the local ``app`` file. In app bundles created from text,
            files paths are flattened to include only the file name.
        """
        if files is None:
            return self

        if isinstance(files, (str, Path)):
            files = [files]

        for file in files or []:
            if self._app_path is not None and Path(file) == self._app_path:
                continue
            self.add_file(file)

        return self

    def add_dir(
        self,
        dir: str | Path,
        flatten: bool = False,
        overwrite: bool = False,
    ) -> ShinyliveApp:
        """
        Add all files in a directory to the Shinylive application.

        Parameters
        ----------
        dir
            The directory to add to the application.
        flatten
            Whether or not to flatten the directory structure. Defaults to ``False``.
            When ``True``, all files are added to the root directory of the application,
            otherwise all files are added into a directory with the same name as the
            input ``dir``.
        overwrite
            Whether or not to overwrite an existing file with the same name. Defaults
            to ``False``.
        """
        dir = Path(dir)
        if not dir.is_dir():
            raise ValueError(f"Directory '{dir}' does not exist or is not a directory.")

        for file in listdir_recursive(dir):
            if flatten:
                # The contents of the directory are added into the "root" of the app
                # bundle, e.g. `"../some_dir/www"` adds its contents to the bundle,
                # i.e. the bundle will have `styles.css` and `app.js`, etc.
                name = str(Path(file).relative_to(dir))
            else:
                # The directory is added into the bundle, e.g. `"../some_dir/www"`
                # is added as `www/` in the bundle.
                name = os.path.join(
                    os.path.basename(dir),
                    str(Path(file).relative_to(dir)),
                )
            self.add_file(file, name, overwrite=overwrite)

        return self

    def add_file_contents(self, file_contents: dict[str, str]) -> ShinyliveApp:
        """
        Directly adds a text file to the Shinylive app.

        Parameters
        ----------
        file_contents
            A dictionary of file names and file contents.
        """
        for file in file_contents:
            self._bundle.append(
                {
                    "name": file,
                    "content": file_contents[file],
                }
            )

        return self

    def add_file(
        self,
        file: str | Path,
        name: Optional[str | Path] = None,
        overwrite: bool = False,
    ) -> ShinyliveApp:
        """
        Add a file to the Shinylive application.

        Parameters
        ----------
        file
            File or directory path to include in the application. On shinylive, this
            file will be stored relative to the main ``app`` file. All files should be
            contained in the same directory as or a subdirectory of the main ``app`` file.
        name
            The name of the file to be used in the app. If not provided, the file name
            will be used, using the relative path from the main ``app`` file if the
            ``ShinyliveIoApp`` was created from local files.
        overwrite
            Whether or not to overwrite an existing file with the same name. Defaults
            to ``False``.
        """
        file_new = read_file(file, self._root_dir)
        if name is not None:
            file_new["name"] = str(name)

        file_names = [file["name"] for file in self._bundle]

        if any([name == file_new["name"] for name in file_names]):
            if overwrite:
                index = file_names.index(file_new["name"])
                self._bundle[index] = file_new
            else:
                raise ValueError(
                    f"File '{file_new['name']}' already exists in app bundle and `overwrite` was `False`."
                )

        self._bundle.append(file_new)

        return self

    def remove_file(self, file: str | Path) -> ShinyliveApp:
        file_names = [file["name"] for file in self._bundle]
        index = None

        if file in file_names:
            # find the index of the file to remove
            index = file_names.index(file)

        if self._root_dir is not None:
            root_dir = self._root_dir.absolute()

            other_path = str(Path(file).absolute().relative_to(root_dir))
            if other_path in file_names:
                index = file_names.index(other_path)

        if index is None:
            raise ValueError(f"File '{file}' not found in app bundle.")

        return self

    def __add__(self, other: str | Path) -> ShinyliveApp:
        other = Path(other)
        new: ShinyliveApp = copy.deepcopy(self)
        if other.is_dir():
            new.add_dir(other)
        else:
            new.add_file(other)
        return new

    def __sub__(self, other: str | Path) -> ShinyliveApp:
        new: ShinyliveApp = copy.deepcopy(self)
        new.remove_file(other)
        return new


def url_encode(
    app: str | Path,
    files: Optional[str | Path | Sequence[str | Path]] = None,
    language: Optional[Literal["py", "r"]] = None,
    mode: Literal["editor", "app"] = "editor",
    header: bool = True,
    host: str = "https://shinylive.io",
) -> str:
    """
    Generate a URL for a [Shinylive application](https://shinylive.io).

    Parameters
    ----------
    app
        The main app file of the Shinylive application. This file should be a Python
        `app.py` or an R `app.R`, `ui.R`, or `server.R` file. This file will be renamed
         `app.py` or `app.R` for shinylive, unless it's named `ui.R` or `server.R`.
    files
        File(s) or directory path(s) to include in the application. On shinylive, these
        files will be stored relative to the main `app` file.
    mode
        The mode of the application, either "editor" or "app". Defaults to "editor".
    language
        The language of the application, or None to autodetect the language. Defaults to
        None.
    header
        Whether to include a header bar in the UI. This is used only if ``mode`` is
        "app". Defaults to True.


    Returns
    -------
        A ShinyliveApp object. Use the `.url()` method to retrieve the Shinylive URL.
    """

    if language is not None and language not in ["py", "r"]:
        raise ValueError(f"Invalid language '{language}', must be either 'py' or 'r'.")

    lang = language if language is not None else detect_app_language(app)

    if isinstance(app, str) and "\n" in app:
        sl_app = ShinyliveApp.from_text(
            app, files, lang, mode=mode, header=header, host=host
        )
    else:
        sl_app = ShinyliveApp.from_local(
            app, files, lang, mode=mode, header=header, host=host
        )

    return sl_app.to_url()


def url_decode(url: str) -> ShinyliveApp:
    """
    Decode a Shinylive URL into a ShinyliveIoApp object.

    Parameters
    ----------
    url
        The Shinylive URL to decode.

    Returns
    -------
        A ShinyliveIoApp object.
    """
    return ShinyliveApp.from_url(url)


def bundle_from_url(url: str) -> list[FileContentJson]:
    from lzstring import LZString  # type: ignore[reportMissingTypeStubs]

    url = url.strip()

    try:
        bundle_json = cast(
            str,
            LZString.decompressFromEncodedURIComponent(  # type: ignore
                url.split("code=")[1]
            ),
        )
        bundle = json.loads(bundle_json)
    except Exception:
        raise ValueError("Could not parse and decode the Shinylive URL code payload.")

    ret: list[FileContentJson] = []

    # bundle should be an array of FileContentJson objects, otherwise raise an error
    if not isinstance(bundle, list):
        raise ValueError(
            "The Shinylive URL was not formatted correctly: `code` did not decode to a list."
        )

    for file in bundle:  # type: ignore
        if not isinstance(file, dict):
            raise ValueError(
                "Invalid Shinylive URL: `code` did not decode to a list of dictionaries."
            )
        if not all(key in file for key in ["name", "content"]):
            raise ValueError(
                "Invalid Shinylive URL: `code` included an object that was missing required fields `name` or `content`."
            )

        for key in ["name", "content"]:
            if not isinstance(file[key], str):
                raise ValueError(
                    f"Invalid Shinylive URL: encoded file bundle contains an file where `{key}` was not a string."
                )

        fc: FileContentJson = {
            "name": file["name"],
            "content": file["content"],
        }

        if "type" in file:
            if file["type"] == "binary":
                fc["type"] = "binary"
            elif file["type"] == "text":
                pass
            else:
                raise ValueError(
                    f"Invalid Shinylive URL: unexpected file type '{file['type']}' in '{file['name']}'."
                )

        if not all(isinstance(value, str) for value in file.values()):  # type: ignore
            raise ValueError(
                f"Invalid Shinylive URL: not all items in '{file['name']}' were strings."
            )
        ret.append(fc)

    return ret


def detect_app_language(app: str | Path) -> Literal["py", "r"]:
    err_not_detected = """
    Could not automatically detect the language of the app. Please specify `language`."""

    if isinstance(app, str) and "\n" in app:
        if re.search(r"^(import|from) shiny", app, re.MULTILINE):
            return "py"
        elif re.search(r"^library\(shiny\)", app, re.MULTILINE):
            return "r"
        else:
            raise ValueError(err_not_detected)

    app = Path(app)

    if app.suffix.lower() == ".py":
        return "py"
    elif app.suffix.lower() == ".r":
        return "r"
    else:
        raise ValueError(err_not_detected)


def listdir_recursive(dir: str | Path) -> list[str]:
    """
    List files in a directory, recursively. Ignores directories or files that start with
    "." or "__".
    """
    dir = Path(dir)
    all_files: list[str] = []

    for root, dirs, files in os.walk(dir):
        # Exclude files and directories that start with `.` or `__`
        dirs[:] = [d for d in dirs if not (d.startswith(".") or d.startswith("__"))]
        files[:] = [f for f in files if not (f.startswith(".") or f.startswith("__"))]

        for file in files:
            if not (file.startswith(".") or file.startswith("__")):
                all_files.append(os.path.join(root, file))

    return all_files


# Copied from https://github.com/posit-dev/py-shiny/blob/main/docs/_renderer.py#L231
def read_file(file: str | Path, root_dir: str | Path | None = None) -> FileContentJson:
    file = Path(file)

    type: Literal["text", "binary"] = "text"

    try:
        with open(file, "r") as f:
            file_content = f.read()
            type = "text"
    except UnicodeDecodeError:
        # If text failed, try binary.
        with open(file, "rb") as f:
            file_content_bin = f.read()
            file_content = base64.b64encode(file_content_bin).decode("utf-8")
            type = "binary"

    file_name = str(file.relative_to(root_dir)) if root_dir else file.name

    return {
        "name": file_name,
        "content": file_content,
        "type": type,
    }


def lzstring_file_bundle(file_bundle: list[FileContentJson]) -> str:
    from lzstring import LZString  # type: ignore[reportMissingTypeStubs]

    file_json = json.dumps(file_bundle)
    file_lz = LZString.compressToEncodedURIComponent(file_json)  # type: ignore[reportUnknownMemberType]
    return cast(str, file_lz)
